#include "flsa.hpp"

namespace FLSA {

using namespace Util;

void UpdMsg(Msg * msg, RealType lin, RealType quad, RealType lambda2,
            RealType * bp) {
  vector<MsgElt>& buf = msg->buf_;
  int buf_sz = buf.size();
  int last_elt = msg->start_idx_ + msg->len_ - 1;

  RealType lin_left = lin, quad_left = quad;
  int new_knot_start = -3;

  for (int knot_idx = msg->start_idx_, m = msg->len_; m; --m, ++knot_idx) {
    if (knot_idx == last_elt) {
      error("numerical error.  failed to find back-pointer segment\n");
    }

    const MsgElt& k = buf[knot_idx];
    RealType x1 = buf[knot_idx + 1].x_;
    if (k.sgn_) {
      lin_left += k.lin_;
      quad_left += k.quad_;
    } else {
      lin_left -= k.lin_;
      quad_left -= k.quad_;
    }

    RealType hit_x = (lambda2 - lin_left) / (2.0 * quad_left);
    if (hit_x < x1) {
      // place a knot here
      new_knot_start = knot_idx - 1;

      if (new_knot_start < 0) {
        error("knot index out of range\n");
      }

      bp[0] = hit_x;
      break;
    }
  }

  RealType lin_right = lin, quad_right = quad;
  int new_knot_end = -2;
  int end_idx = msg->start_idx_ + msg->len_ - 1;

  double neg_lam2 = -lambda2;

  for (int knot_idx = end_idx, m = msg->len_; m; --m, --knot_idx) {
    if (knot_idx == msg->start_idx_) {
      error("numerical error.  failed to find back-pointer segment\n");
    }
    const MsgElt& k = buf[knot_idx];
    RealType x1 = buf[knot_idx - 1].x_;

    if (k.sgn_) {
      lin_right -= k.lin_;
      quad_right -= k.quad_;
    } else {
      lin_right += k.lin_;
      quad_right += k.quad_;
    }

    RealType hit_x = (neg_lam2 - lin_right) / (2.0 * quad_right);
    if (hit_x > x1) {
      // place a knot here
      new_knot_end = knot_idx + 1;
      bp[1] = hit_x;
      break;
    }
  }

  // update the message and add the new knots
  MsgElt * k0 = &buf[new_knot_start];
  k0->x_ = R_NegInf;
  k0->sgn_ = true;
  k0->lin_ = lambda2;
  k0->quad_ = 0.0;
  msg->init_knot_ = *k0;

  k0 = &buf[new_knot_start + 1];
  k0->x_ = bp[0];
  k0->sgn_ = true;
  k0->lin_ = lin_left - lambda2;
  k0->quad_ = quad_left;

  k0 = &buf[new_knot_end];
  k0->x_ = R_PosInf;
  k0->sgn_ = false;
  k0->lin_ = -lambda2;
  k0->quad_ = 0.0;
  msg->end_knot_ = *k0;

  k0 = &buf[new_knot_end-1];
  k0->x_ = bp[1];
  k0->sgn_ = false;
  k0->lin_ = lin_right + lambda2;
  k0->quad_ = quad_right;

  msg->start_idx_ = new_knot_start;
  msg->len_ = 1 + new_knot_end - new_knot_start;
}

void UpdMsgOpt(Msg * msg, RealType lambda2, RealType lin, RealType quad,
               RealType * bp) {
  // Assumes that msg->buf_[0] and msg->buf_[1] contain the end-point knots and
  // that we do not need to create them

  vector<MsgElt>& buf = msg->buf_;
  //RealType quad = -0.5;

  buf[msg->start_idx_].x_ = R_NegInf;
  buf[msg->start_idx_ + msg->len_ - 1].x_ = R_PosInf;

  RealType lin_left = lin, quad_left = quad;
  int new_knot_start = -3;

  int knot_idx = msg->start_idx_;
  const MsgElt * k = &(msg->init_knot_);
  RealType x1 = buf[knot_idx + 1].x_;
  if (k->sgn_) {
    lin_left += k->lin_;
    quad_left += k->quad_;
  } else {
    lin_left -= k->lin_;
    quad_left -= k->quad_;
  }

  RealType hit_x = (lambda2 - lin_left) / (2.0 * quad_left);
  if (hit_x < x1) {
    // place a knot here
    new_knot_start = knot_idx - 1;

    bp[0] = hit_x;
  } else {
    ++knot_idx;
    k = &buf[knot_idx];
    for (; ; ++knot_idx, ++k) {
      x1 = k[1].x_;
      if (k->sgn_) {
        lin_left += k->lin_;
        quad_left += k->quad_;
      } else {
        lin_left -= k->lin_;
        quad_left -= k->quad_;
      }

      hit_x = (lambda2 - lin_left) / (2.0 * quad_left);
      if (hit_x < x1) {
        // place a knot here
        new_knot_start = knot_idx - 1;

        bp[0] = hit_x;
        break;
      }
    }
  }

  RealType lin_right = lin, quad_right = quad;
  int new_knot_end = -2;
  int end_idx = msg->start_idx_ + msg->len_ - 1;

  double neg_lam2 = -lambda2;

  knot_idx = end_idx;
  k = &(msg->end_knot_);
  x1 = buf[knot_idx-1].x_;

  if (k->sgn_) {
    lin_right -= k->lin_;
    quad_right -= k->quad_;
  } else {
    lin_right += k->lin_;
    quad_right += k->quad_;
  }

  hit_x = (neg_lam2 - lin_right) / (2.0 * quad_right);
  if (hit_x > x1) {
    // place a knot here
    new_knot_end = knot_idx + 1;
    bp[1] = hit_x;
  } else {
    --knot_idx;
    k = &buf[knot_idx];
    for (; ; --knot_idx, --k) {
      x1 = k[-1].x_;

      if (k->sgn_) {
        lin_right -= k->lin_;
        quad_right -= k->quad_;
      } else {
        lin_right += k->lin_;
        quad_right += k->quad_;
      }

      hit_x = (neg_lam2 - lin_right) / (2.0 * quad_right);
      if (hit_x > x1) {
        // place a knot here
        new_knot_end = knot_idx + 1;
        bp[1] = hit_x;
        break;
      }
    }
  }

  MsgElt * k0 = &buf[new_knot_start + 1];
  k0->x_ = bp[0];
  k0->sgn_ = true;
  k0->lin_ = lin_left - lambda2;
  k0->quad_ = quad_left;

  k0 = &buf[new_knot_end-1];
  k0->x_ = bp[1];
  k0->sgn_ = false;
  k0->lin_ = lin_right + lambda2;
  k0->quad_ = quad_right;

  msg->start_idx_ = new_knot_start;
  msg->len_ = 1 + new_knot_end - new_knot_start;
}

double MaxMsg(const Msg& msg, double * max_val) {
  const vector<MsgElt>& buf = msg.buf_;
  int buf_sz = buf.size();
  RealType lin_left = 0.0, quad_left = 0.0;

  int last_idx = msg.start_idx_ + msg.len_ - 1;

  for (int knot_idx = msg.start_idx_, m = msg.len_; m; --m, ++knot_idx) {
    bool end_knot = (knot_idx == last_idx);
    const MsgElt& k = (knot_idx == msg.start_idx_) ?
                       msg.init_knot_ :
                       (end_knot ? msg.end_knot_ : buf[knot_idx]);
    //double x0 = k.x_;
    RealType x1 = (knot_idx == last_idx-1) ? msg.end_knot_.x_ : buf[knot_idx + 1].x_;

    if (k.sgn_) {
      lin_left += k.lin_;
      quad_left += k.quad_;
    } else {
      lin_left -= k.lin_;
      quad_left -= k.quad_;
    }

    if (quad_left == 0.0) {
      continue;
    }

    RealType hit_x = -lin_left / (2.0 * quad_left);
    if (hit_x < x1) {
      if (max_val) {
        *max_val = hit_x * lin_left + hit_x * hit_x * quad_left;
      }
      return(hit_x);
    }
  }

  error("FLSA::MaxMsg : failed to maximize message\n");
  return(R_NegInf);
}

void InitMsg(Msg * msg, int init_sz, RealType lin, RealType quad,
             RealType lambda2) {
  msg->len_ = 2;
  int i = msg->start_idx_ = init_sz / 2;

  msg->buf_ = vector<MsgElt>(init_sz);

  msg->buf_[i].x_ = R_NegInf;
  msg->buf_[i].sgn_ = true;
  msg->buf_[i].lin_ = lin;
  msg->buf_[i].quad_ = quad;

  msg->buf_[i+1].x_ = R_PosInf;
  msg->buf_[i+1].sgn_ = false;
  msg->buf_[i+1].lin_ = lin;
  msg->buf_[i+1].quad_ = quad;

  msg->init_knot_.x_ = R_NegInf;
  msg->init_knot_.sgn_ = true;
  msg->init_knot_.lin_ = lambda2;
  msg->init_knot_.quad_ = 0.0;

  msg->end_knot_.x_ = R_PosInf;
  msg->end_knot_.sgn_ = false;
  msg->end_knot_.lin_ = -lambda2;
  msg->end_knot_.quad_ = 0.0;
}

void ShiftMsg(Msg * msg, int check_freq) {
  if (msg->len_ > msg->buf_.size() - 20 * check_freq) {
    vector<MsgElt> new_buf(msg->buf_.size() * 3);
    vector<MsgElt> * old_buf = &msg->buf_;

    int new_start = (new_buf.size() / 4);
    int old_start = msg->start_idx_;

    for (int k = 0; k < msg->len_; ++k) {
      new_buf[k + new_start] = (*old_buf)[k + old_start];
    }
    msg->buf_.swap(new_buf);
    msg->start_idx_ = new_start;
  }

  if (msg->start_idx_ < 5 * check_freq) {
    int new_start = ((msg->buf_.size() - msg->len_)/2);
    int old_start = msg->start_idx_;
    vector<MsgElt> * buf = &msg->buf_;

    for (int k = msg->len_ - 1; k >= 0; --k) {
      (*buf)[k + new_start] = (*buf)[k + old_start];
    }
    msg->start_idx_ = new_start;

  } else if (msg->start_idx_ + msg->len_ > msg->buf_.size() - 5 * check_freq) {
    int new_start = ((msg->buf_.size() - msg->len_)/2);
    int old_start = msg->start_idx_;
    vector<MsgElt> * buf = &msg->buf_;

    for (int k = 0; k < msg->len_; ++k) {
      (*buf)[k + new_start] = (*buf)[k + old_start];
    }
    msg->start_idx_ = new_start;
  }
}

void BackTrace(double * x_hat, int seq_len, RealType * back_pointers,
               RealType last_msg_max) {
  RealType z = x_hat[seq_len - 1] = last_msg_max;
  double * x0 = x_hat + seq_len - 2;

  RealType * bp = back_pointers + (2 * (seq_len - 2));
  for (int idx = seq_len-1; idx; --idx, bp -= 2, --x0) {
    if (z < bp[0]) {
      z = *x0 = bp[0];
    } else if (z > bp[1]) {
      z = *x0 = bp[1];
    } else {
      *x0 = z;
    }
  }
}

int NumSegments(int seq_len, RealType * back_pointers, RealType last_msg_max) {
  int n_seg = 1;
  RealType z = last_msg_max;
  RealType * bp = back_pointers + (2 * (seq_len - 2));
  for (int idx = seq_len-1; idx > 0; --idx, bp -= 2) {
    if (z < bp[0]) {
      z = bp[0];
      ++n_seg;
    } else if (z > bp[1]) {
      z = bp[1];
      ++n_seg;
    }
  }
  return(n_seg);
}

void SegmentedBackTrace(double * x_hat, int * seg_bds, int seq_len,
                        RealType * back_pointers,
                        RealType last_msg_max, int n_segs) {
  int seg_idx = n_segs - 1;
  seg_bds[2 * seg_idx + 1] = seq_len - 1;
  RealType z = x_hat[seg_idx] = last_msg_max;

  RealType * bp = back_pointers + (2 * (seq_len - 2));
  for (int idx = seq_len-1; idx > 0; --idx, bp -= 2) {
    if (z < bp[0]) {
      seg_bds[2 * seg_idx + 1] = idx;
      seg_idx--;
      seg_bds[2 * seg_idx + 1] = idx-1;
      z = x_hat[seg_idx] = bp[0];
    } else if (z > bp[1]) {
      seg_bds[2 * seg_idx + 1] = idx;
      seg_idx--;
      seg_bds[2 * seg_idx + 1] = idx-1;
      z = x_hat[seg_idx] = bp[1];
    }
  }
  if (seg_idx != 0) {
    error("SegmentedBackTrace : failed\n");
  }
  seg_bds[0] = 0;
}

void SoftThresh(double * x, int len, double thresh) {
  double neg_thresh = -thresh;
  for (; len; --len, ++x) {
    if (*x > thresh) {
      *x -= thresh;
    } else if (*x < neg_thresh) {
      *x += thresh;
    } else {
      *x = 0.0;
    }
  }
}

}


extern "C" {
SEXP FLSA_RI(SEXP inp_seq_sxp, SEXP lambda2_sxp, SEXP lambda1_sxp,
             SEXP fit_sxp, SEXP init_buf_sz_sxp) {
  using namespace Util;

  double * x = REAL(inp_seq_sxp);
  int seq_len = LENGTH(inp_seq_sxp) / 2;

  double * lambda2 = REAL(lambda2_sxp);
  int lamda2_len = LENGTH(lambda2_sxp);
  double lambda1 = GetNumeric(lambda1_sxp, 0.0, NULL);

  if (seq_len < 2) {
    error("FLSA_RI : input vector has length less than two\n");
  }
  FLSA::RealType * back_pointers =
      (FLSA::RealType*)malloc(seq_len * 2 * sizeof(FLSA::RealType));

  int check_freq = 40;
  int init_buf_sz = GetInt(init_buf_sz_sxp, 100, NULL);

  if (init_buf_sz < 30 * check_freq) {
    error("FLSA_RI : initial buffer size is too small\n");
  }

  FLSA::Msg msg;
  FLSA::InitMsg(&msg, init_buf_sz, 0.0, 0.0, lambda2[0]);

  int max_msg_sz = 0;

  double * bp = back_pointers + 2;
  double * x0 = x;
  for (int j = 0; j < seq_len; ++j, bp += 2, x0 += 2) {
    double lam2 = lambda2[j % lamda2_len];
    if (lam2 <= 0.0) {
      error("FLSA_RI : lambda2's must be strictly positive\n");
    }
    FLSA::UpdMsg(&msg, x[0], x[1], lam2, bp);
    if (j % check_freq == 0) FLSA::ShiftMsg(&msg, check_freq);
  }

  FLSA::RealType last_msg_max = FLSA::MaxMsg(msg, NULL);

  if (IS_NUMERIC(fit_sxp)) {
    if (seq_len > LENGTH(fit_sxp)) {
      error("FLSA_RI : output vector wrong length\n");
    }
    double * x_hat = REAL(fit_sxp);
    FLSA::BackTrace(x_hat, seq_len, back_pointers, last_msg_max);

    if (lambda1 > 0.0) {
      FLSA::SoftThresh(x_hat, seq_len, lambda1);
    }
  } else {
    int n_segs = FLSA::NumSegments(seq_len, back_pointers, last_msg_max);

    SEXP seg_bds_sxp = NEW_INTEGER(n_segs * 2);
    SET_VECTOR_ELT(fit_sxp, 0, seg_bds_sxp);
    int * seg_bds = INTEGER(VECTOR_ELT(fit_sxp, 0));

    SEXP seg_vals_sxp = NEW_NUMERIC(n_segs);
    SET_VECTOR_ELT(fit_sxp, 1, seg_vals_sxp);
    double * x_hat = REAL(VECTOR_ELT(fit_sxp, 1));

    FLSA::SegmentedBackTrace(x_hat, seg_bds, seq_len, back_pointers,
                             last_msg_max, n_segs);

    if (lambda1 > 0.0) {
      FLSA::SoftThresh(x_hat, n_segs, lambda1);
    }
  }

  free(back_pointers);

  return(R_NilValue);
}


SEXP FLSAOpt_RI(SEXP inp_seq_sxp, SEXP lambda2_sxp, SEXP lambda1_sxp,
                SEXP fit_sxp, SEXP init_buf_sz_sxp, SEXP obs_wts_sxp) {
  using namespace Util;

  double * x = REAL(inp_seq_sxp);
  int seq_len = LENGTH(inp_seq_sxp);

  double lambda2 = GetNumeric(lambda2_sxp, 1.0, NULL);
  double lambda1 = GetNumeric(lambda1_sxp, 0.0, NULL);

  if (lambda2 <= 0.0) {
    error("FLSA_RI : lambda2's must be strictly positive\n");
  }

  if (seq_len < 2) {
    error("FLSA_RI : input vector has length less than two\n");
  }
  FLSA::RealType * back_pointers =
      (FLSA::RealType*)malloc(seq_len * 2 * sizeof(FLSA::RealType));

  int check_freq = 40;
  int init_buf_sz = GetInt(init_buf_sz_sxp, 100, NULL);

  if (init_buf_sz < 30 * check_freq) {
    error("FLSA_RI : initial buffer size is too small\n");
  }

  FLSA::Msg msg;
  FLSA::InitMsg(&msg, init_buf_sz, 0.0, 0.0, lambda2);

  int max_msg_sz = 0;

  double * bp = back_pointers + 2;
  double * x0 = x + 1;
  int check_msg = check_freq - 1;

  if (obs_wts_sxp == R_NilValue) {
    FLSA::UpdMsg(&msg, x[0], -0.5, lambda2, back_pointers);
    for (int j = 1; j < seq_len; ++j, bp += 2, ++x0, --check_msg) {
      FLSA::UpdMsgOpt(&msg, lambda2, *x0, -0.5, bp);
      if (!check_msg) {
        check_msg = check_freq - 1;
        FLSA::ShiftMsg(&msg, check_freq);
      }
    }
  } else {
    if (!IS_NUMERIC(obs_wts_sxp) || LENGTH(obs_wts_sxp) < seq_len) {
      error("FLSAOpt_RI : obs_wts_sxp is invalid\n");
    }

    double * obs_wts = REAL(obs_wts_sxp);
    FLSA::UpdMsg(&msg, x[0] * obs_wts[0], -0.5 * obs_wts[0], lambda2,
                 back_pointers);
    double * wt = obs_wts + 1;
    for (int j = 1; j < seq_len; ++j, bp += 2, ++x0, ++wt, --check_msg) {
      FLSA::UpdMsgOpt(&msg, lambda2, (*x0) * (*wt), -0.5 * (*wt), bp);
      if (!check_msg) {
        check_msg = check_freq - 1;
        FLSA::ShiftMsg(&msg, check_freq);
      }
    }
  }

  FLSA::RealType last_msg_max = FLSA::MaxMsg(msg, NULL);

  if (IS_NUMERIC(fit_sxp)) {
    if (seq_len > LENGTH(fit_sxp)) {
      error("FLSA_RI : output vector wrong length\n");
    }
    double * x_hat = REAL(fit_sxp);
    FLSA::BackTrace(x_hat, seq_len, back_pointers, last_msg_max);

    if (lambda1 > 0.0) {
      FLSA::SoftThresh(x_hat, seq_len, lambda1);
    }
  } else {
    int n_segs = FLSA::NumSegments(seq_len, back_pointers, last_msg_max);

    SEXP seg_bds_sxp = NEW_INTEGER(n_segs * 2);
    SET_VECTOR_ELT(fit_sxp, 0, seg_bds_sxp);
    int * seg_bds = INTEGER(VECTOR_ELT(fit_sxp, 0));

    SEXP seg_vals_sxp = NEW_NUMERIC(n_segs);
    SET_VECTOR_ELT(fit_sxp, 1, seg_vals_sxp);
    double * x_hat = REAL(VECTOR_ELT(fit_sxp, 1));

    FLSA::SegmentedBackTrace(x_hat, seg_bds, seq_len, back_pointers,
                             last_msg_max, n_segs);

    if (lambda1 > 0.0) {
      FLSA::SoftThresh(x_hat, n_segs, lambda1);
    }
  }

  free(back_pointers);

  return(R_NilValue);
}  // SEXP BinaryMatXXT_RI

}  // extern "C"
