#include "util.hpp"
#include "flsa.hpp"

// This code modifies the FLDP algorithm to perform isotonic and unimodal
// regression.

namespace IsotonicDp {

typedef FLSA::RealType RealType;

using namespace Util;

struct MsgElt {
  // the location of the knot
  RealType x_;
  // the sign variable that tells us
  // whether this was a left or right end-point of the
  // segment
  bool sgn_;
  // a delta which can be used to reconstruct the function
  // if we move from the first knot to the last or from
  // the last to the first
  RealType lin_;
  RealType quad_;
  RealType const_;

  void ToFLSAMsgElt(FLSA::MsgElt * m) const {
    m->x_ = x_;
    m->sgn_ = sgn_;
    m->lin_ = lin_;
    m->quad_ = quad_;
  }
};

class Msg {
public:
  vector<MsgElt> buf_;
  int start_idx_;
  int len_;
  MsgElt init_knot_;
  MsgElt end_knot_;

  void ToFLSAMsg(FLSA::Msg * m) const {
    m->buf_ = vector<FLSA::MsgElt>(buf_.size());
    m->start_idx_ = start_idx_;
    m->len_ = len_;
    init_knot_.ToFLSAMsgElt(&m->init_knot_);
    end_knot_.ToFLSAMsgElt(&m->end_knot_);
    for (int i = 0, sz = buf_.size(); i < sz; ++i) {
      buf_[i].ToFLSAMsgElt(&m->buf_[i]);
    }
  }
};

double MaxMsg(const Msg& msg, double * max_val) {
  FLSA::Msg m;
  msg.ToFLSAMsg(&m);
  return(FLSA::MaxMsg(m, max_val));
}

Msg * AllocMsgCopy(const Msg& msg) {
  Msg * r = new Msg;
  r->buf_ = vector<MsgElt>(msg.len_);

  r->start_idx_ = 0;
  r->init_knot_ = msg.init_knot_;
  r->end_knot_ = msg.end_knot_;
  r->len_ = msg.len_;

  for (int i = 0, j = msg.start_idx_; i < msg.len_; ++i, ++j) {
    r->buf_[i] = msg.buf_[j];
  }
  return(r);
}

void ShiftMsg(Msg * msg, int check_freq) {
  if (msg->len_ > msg->buf_.size() - 20 * check_freq) {
    vector<MsgElt> new_buf(msg->buf_.size() * 3);
    vector<MsgElt> * old_buf = &msg->buf_;

    int new_start = (new_buf.size() / 4);
    int old_start = msg->start_idx_;

    for (int k = 0; k < msg->len_; ++k) {
      new_buf[k + new_start] = (*old_buf)[k + old_start];
    }
    msg->buf_.swap(new_buf);
    msg->start_idx_ = new_start;
  }

  if (msg->start_idx_ < 5 * check_freq) {
    //Rprintf("shift message 2\n");
    int new_start = ((msg->buf_.size() - msg->len_)/2);
    int old_start = msg->start_idx_;
    vector<MsgElt> * buf = &msg->buf_;

    for (int k = msg->len_ - 1; k >= 0; --k) {
      (*buf)[k + new_start] = (*buf)[k + old_start];
    }
    msg->start_idx_ = new_start;

  } else if (msg->start_idx_ + msg->len_ > msg->buf_.size() - 5 * check_freq) {
    //Rprintf("shift message 3\n");
    int new_start = ((msg->buf_.size() - msg->len_)/2);
    int old_start = msg->start_idx_;
    vector<MsgElt> * buf = &msg->buf_;

    for (int k = 0; k < msg->len_; ++k) {
      (*buf)[k + new_start] = (*buf)[k + old_start];
    }
    msg->start_idx_ = new_start;
  }
}

void FirstMsg(Msg * msg, int init_sz, RealType lin, RealType quad,
              RealType lambda2, RealType * bp) {
  msg->len_ = 3;
  msg->start_idx_ = init_sz / 2;

  msg->buf_ = vector<MsgElt>(init_sz);

  vector<MsgElt>& buf = msg->buf_;
  int buf_sz = buf.size();
  bp[0] = R_NegInf;
  bp[1] = (-lambda2 - lin) / (2.0 * quad);

  int new_knot_start = msg->start_idx_;

  // update the message and add the new knots
  MsgElt * k0 = &buf[msg->start_idx_];
  k0->x_ = R_NegInf;
  k0->sgn_ = true;
  k0->lin_ = lin;
  k0->quad_ = quad;
  k0->const_ = 0.0;
  msg->init_knot_ = *k0;

  RealType const_shift = bp[1] * lin + bp[1] * bp[1] * quad + lambda2 * bp[1];
  k0 = &buf[msg->start_idx_+2];
  k0->x_ = R_PosInf;
  k0->sgn_ = false;
  k0->lin_ = -lambda2;
  k0->quad_ = 0.0;
  k0->const_ = const_shift;
  msg->end_knot_ = *k0;


  k0 = &buf[msg->start_idx_+1];
  k0->x_ = bp[1];
  k0->sgn_ = false;
  k0->lin_ = lin + lambda2;
  k0->const_ = 0.0;
  k0->quad_ = quad;

  msg->len_ = 3;
}

void UpdMsgOpt(Msg * msg, RealType lin, RealType quad, RealType lambda2,
               RealType * bp) {
  // Assumes that msg->buf_[0] and msg->buf_[1] contain the end-point knots and
  // that we do not need to create them.

  vector<MsgElt>& buf = msg->buf_;

  buf[msg->start_idx_].x_ = R_NegInf;
  buf[msg->start_idx_ + msg->len_ - 1].x_ = R_PosInf;

  bp[0] = R_NegInf;

  RealType lin_right = lin, quad_right = quad;
  int new_knot_end = -2;
  int end_idx = msg->start_idx_ + msg->len_ - 1;

  double neg_lam2 = -lambda2;

  int start_idx = msg->start_idx_;
  int knot_idx = end_idx;
  MsgElt * k = &(msg->end_knot_);
  double x1 = (knot_idx == start_idx+1)
              ? msg->init_knot_.x_
              : buf[knot_idx-1].x_;
  double cur_const = msg->end_knot_.const_;

  if (k->sgn_) {
    lin_right -= k->lin_;
    quad_right -= k->quad_;
  } else {
    lin_right += k->lin_;
    quad_right += k->quad_;
  }

  double hit_x = (neg_lam2 - lin_right) / (2.0 * quad_right);
  if (hit_x > x1) {
    // place a knot here
    new_knot_end = knot_idx + 1;
    bp[1] = hit_x;
  } else {
    --knot_idx;

    for (k = &buf[knot_idx]; ; --knot_idx, --k) {
      x1 = (knot_idx == start_idx+1) ? msg->init_knot_.x_ : k[-1].x_;

      if (k->sgn_) {
        lin_right -= k->lin_;
        quad_right -= k->quad_;
      } else {
        lin_right += k->lin_;
        quad_right += k->quad_;
      }

      hit_x = (neg_lam2 - lin_right) / (2.0 * quad_right);
      if (hit_x > x1) {
        // place a knot here
        new_knot_end = knot_idx + 1;
        bp[1] = hit_x;
        cur_const = k->const_;
        break;
      }
    }
  }

  msg->init_knot_.lin_ += lin;
  msg->init_knot_.quad_ += quad;

  MsgElt * k0 = &buf[new_knot_end-1];
  k0->x_ = bp[1];
  k0->sgn_ = false;
  k0->lin_ = lin_right + lambda2;
  k0->quad_ = quad_right;
  k0->const_ = cur_const;

  msg->end_knot_.const_ = cur_const + lambda2 * bp[1] + bp[1] * lin_right
                          + bp[1] * bp[1] * quad_right;
  RealType const_shift = bp[1] * lin + bp[1] * bp[1] * quad + lambda2 * bp[1];

  msg->len_ = 1 + new_knot_end - msg->start_idx_;
}

double MaxMsgProduct(const Msg& msg1, const Msg& msg2, RealType lin,
                     RealType quad, double * max_val) {
  const vector<MsgElt>& buf1 = msg1.buf_;
  const vector<MsgElt>& buf2 = msg2.buf_;
  int buf_sz1 = buf1.size();
  int buf_sz2 = buf2.size();
  RealType lin_left1 = msg1.init_knot_.lin_, quad_left1 = msg1.init_knot_.quad_;
  RealType lin_left2 = msg2.init_knot_.lin_, quad_left2 = msg2.init_knot_.quad_;

  int last_idx1 = msg1.start_idx_ + msg1.len_ - 1;
  int last_idx2 = msg2.start_idx_ + msg2.len_ - 1;

  int knot_idx1 = msg1.start_idx_ + 1;
  int knot_idx2 = msg2.start_idx_ + 1;

  double cur_const1 = 0.0;
  double cur_const2 = 0.0;

  while (knot_idx1 <= last_idx1 || knot_idx2 <= last_idx2) {
    if (knot_idx1 >= buf1.size() || knot_idx2 >= buf2.size()) {
      error("knot_idx out of range\n");
    }
    bool end_knot1 = (knot_idx1 >= last_idx1);
    bool end_knot2 = (knot_idx2 >= last_idx2);
    const MsgElt& k1 = end_knot1 ? msg1.end_knot_ : buf1[knot_idx1];
    const MsgElt& k2 = end_knot2 ? msg2.end_knot_ : buf2[knot_idx2];

    RealType x1 = (k1.x_ < k2.x_) ? k1.x_ : k2.x_;

    // see if there is a hit
    if (quad_left1 == 0.0 && quad_left2 == 0.0 && quad == 0.0) {
      continue;
    }

    RealType hit_x = -(lin + lin_left1 + lin_left2)
                     / (2.0 * (quad + quad_left1 + quad_left2));
    if (hit_x < x1) {
      if (max_val) {
        *max_val = (lin + lin_left1 + lin_left2) * hit_x
                   + hit_x * hit_x * (quad + quad_left1 + quad_left2);
        *max_val += k1.const_ + k2.const_;
      }
      return(hit_x);
    }

    // increment the coefficients and move to the next knot
    if (k1.x_ == k2.x_) {
      if (k1.sgn_) {
        lin_left1 += k1.lin_;
        quad_left1 += k1.quad_;
      } else {
        lin_left1 -= k1.lin_;
        quad_left1 -= k1.quad_;
      }
      ++knot_idx1;

      if (k2.sgn_) {
        lin_left2 += k2.lin_;
        quad_left2 += k2.quad_;
      } else {
        lin_left2 -= k2.lin_;
        quad_left2 -= k2.quad_;
      }
      ++knot_idx2;

    } else if (k1.x_ <= k2.x_) {
      if (k1.sgn_) {
        lin_left1 += k1.lin_;
        quad_left1 += k1.quad_;
      } else {
        lin_left1 -= k1.lin_;
        quad_left1 -= k1.quad_;
      }
      ++knot_idx1;
    } else {
      if (k2.sgn_) {
        lin_left2 += k2.lin_;
        quad_left2 += k2.quad_;
      } else {
        lin_left2 -= k2.lin_;
        quad_left2 -= k2.quad_;
      }
      ++knot_idx2;
    }
  }

  error("FLSA::MaxMsg : failed to maximize message\n");
  return(R_NegInf);
}

}


extern "C" {

SEXP IsotonicDpOpt_RI(SEXP inp_seq_sxp, SEXP lambda2_sxp, SEXP fit_sxp,
                      SEXP init_buf_sz_sxp, SEXP obs_wts_sxp) {
  using namespace Util;
  using namespace IsotonicDp;

  double * x = REAL(inp_seq_sxp);
  int seq_len = LENGTH(inp_seq_sxp);

  double lambda2 = GetNumeric(lambda2_sxp, 1.0, NULL);

  if (lambda2 < 1e-10) {
    error("FLSA_RI : lambda2's must be strictly positive\n");
  }

  if (seq_len < 2) {
    error("FLSA_RI : input vector has length less than two\n");
  }
  RealType * back_pointers = new RealType[seq_len * 2];

  int check_freq = 40;
  int init_buf_sz = GetInt(init_buf_sz_sxp, 100, NULL);

  if (init_buf_sz < 30 * check_freq) {
    error("FLSA_RI : initial buffer size is too small\n");
  }

  double * obs_wts = REAL(obs_wts_sxp);
  Msg msg;
  FirstMsg(&msg, init_buf_sz, x[0] * obs_wts[0], -0.5 * obs_wts[0], lambda2,
           back_pointers);

  int max_msg_sz = 0;

  double * bp = back_pointers + 2;
  double * x0 = x + 1;
  int check_msg = check_freq - 1;

  if (!IS_NUMERIC(obs_wts_sxp) || LENGTH(obs_wts_sxp) < seq_len) {
    error("FLSAOpt_RI : obs_wts_sxp is invalid\n");
  }

  double * wt = obs_wts + 1;
  for (int j = 1; j < seq_len; ++j, bp += 2, ++x0, ++wt, --check_msg) {
    UpdMsgOpt(&msg, (*x0) * (*wt), -0.5 * (*wt), lambda2, bp);
    if (!check_msg) {
      check_msg = check_freq - 1;
      ShiftMsg(&msg, check_freq);
    }
  }

  FLSA::Msg msg_copy;
  msg.ToFLSAMsg(&msg_copy);
  RealType last_msg_max = MaxMsg(msg, NULL);

  if (IS_NUMERIC(fit_sxp)) {
    if (seq_len > LENGTH(fit_sxp)) {
      error("FLSA_RI : output vector wrong length\n");
    }
    double * x_hat = REAL(fit_sxp);
    FLSA::BackTrace(x_hat, seq_len, back_pointers, last_msg_max);
  } else {
    int n_segs = FLSA::NumSegments(seq_len, back_pointers, last_msg_max);

    SEXP seg_bds_sxp = NEW_INTEGER(n_segs * 2);
    SET_VECTOR_ELT(fit_sxp, 0, seg_bds_sxp);
    int * seg_bds = INTEGER(VECTOR_ELT(fit_sxp, 0));

    SEXP seg_vals_sxp = NEW_NUMERIC(n_segs);
    SET_VECTOR_ELT(fit_sxp, 1, seg_vals_sxp);
    double * x_hat = REAL(VECTOR_ELT(fit_sxp, 1));

    FLSA::SegmentedBackTrace(x_hat, seg_bds, seq_len, back_pointers,
                             last_msg_max, n_segs);
  }

  delete back_pointers;

  return(R_NilValue);
}  // SEXP BinaryMatXXT_RI

SEXP UnimodalDpOpt_RI(SEXP inp_seq_sxp, SEXP lambda2_sxp,
                      SEXP mode_crit_sxp, SEXP mode_height_sxp, SEXP fit_sxp,
                      SEXP init_buf_sz_sxp, SEXP obs_wts_sxp) {
  using namespace Util;
  using namespace IsotonicDp;

  double * x = REAL(inp_seq_sxp);
  int seq_len = LENGTH(inp_seq_sxp);

  double lambda2 = GetNumeric(lambda2_sxp, 1.0, NULL);

  if (lambda2 < 1e-10) {
    error("FLSA_RI : lambda2's must be strictly positive\n");
  }

  if (seq_len < 2) {
    error("FLSA_RI : input vector has length less than two\n");
  }
  RealType * back_pointers_rev = new RealType[seq_len * 2];
  RealType * back_pointers_fwd = new RealType[seq_len * 2];

  int check_freq = 40;
  int init_buf_sz = GetInt(init_buf_sz_sxp, 100, NULL);

  if (init_buf_sz < 30 * check_freq) {
    error("FLSA_RI : initial buffer size is too small\n");
  }

  int max_msg_sz = 0;


  double * bp_rev = back_pointers_rev + 2;
  double * bp_fwd = back_pointers_fwd + 2;
  int check_msg = check_freq - 1;

  Msg * null_msg = NULL;
  vector<Msg *> fwd_msgs(seq_len, null_msg);
  vector<Msg *> rev_msgs(seq_len, null_msg);

  if (!IS_NUMERIC(obs_wts_sxp) || LENGTH(obs_wts_sxp) < seq_len) {
    error("FLSAOpt_RI : obs_wts_sxp is invalid\n");
  }

  double * obs_wts = REAL(obs_wts_sxp);

  Msg fwd_msg;

  FirstMsg(&fwd_msg, init_buf_sz, x[0] * obs_wts[0], -0.5 * obs_wts[0], lambda2,
           back_pointers_fwd);
  fwd_msgs[0] = AllocMsgCopy(fwd_msg);

  double * x0 = x + 1;
  double * wt = obs_wts + 1;
  for (int j = 1; j < seq_len; ++j, bp_fwd += 2, ++x0, ++wt, --check_msg) {
    UpdMsgOpt(&fwd_msg, (*x0) * (*wt), -0.5 * (*wt), lambda2, bp_fwd);
    if (!check_msg) {
      check_msg = check_freq - 1;
      ShiftMsg(&fwd_msg, check_freq);
    }
    fwd_msgs[j] = AllocMsgCopy(fwd_msg);
  }

  Msg rev_msg;

  x0 = x + seq_len-1;
  wt = obs_wts + seq_len-1;
  FirstMsg(&rev_msg, init_buf_sz, (*x0) * (*wt), -0.5 * (*wt), lambda2,
           back_pointers_rev);
  rev_msgs[0] = IsotonicDp::AllocMsgCopy(rev_msg);

  --x0;
  --wt;

  for (int j = 1; j < seq_len; ++j, bp_rev += 2, --wt, --x0, --check_msg) {
    UpdMsgOpt(&rev_msg, (*x0) * (*wt), -0.5 * (*wt), lambda2, bp_rev);
    if (!check_msg) {
      check_msg = check_freq - 1;
      ShiftMsg(&rev_msg, check_freq);
    }
    rev_msgs[j] = IsotonicDp::AllocMsgCopy(rev_msg);
  }

  double * mode_crit = REAL(mode_crit_sxp);
  double * mode_height = REAL(mode_height_sxp);

  for (int j = 1; j+1 < seq_len; ++j) {
    mode_height[j] = IsotonicDp::MaxMsgProduct(
        *fwd_msgs[j-1], *rev_msgs[seq_len-j-2], obs_wts[j] * x[j],
        -0.5 * obs_wts[j], mode_crit + j);
  }

  mode_crit[0] = mode_crit[seq_len - 1] = R_NegInf;
  mode_height[0] = R_NegInf;
  mode_height[seq_len - 1] = R_NegInf;

  int opt_idx = -1;
  double opt_val = R_NegInf;
  for (int j = 0; j < seq_len; ++j) {
    if (opt_val < mode_crit[j]) {
      opt_val = mode_crit[j];
      opt_idx = j;
    }
  }

  if (opt_idx == 0) {

  } else if (opt_idx == seq_len - 1) {

  } else {

    double * tmp_fit = new double[seq_len];
    double * x_hat = REAL(fit_sxp);
    FLSA::BackTrace(tmp_fit, seq_len-opt_idx, back_pointers_rev, mode_height[opt_idx]);
    for (int j = opt_idx, i = seq_len-opt_idx-1; j < seq_len; ++j, --i) {
      x_hat[j] = tmp_fit[i];
    }
    delete tmp_fit;

    FLSA::BackTrace(x_hat, opt_idx+1, back_pointers_fwd, mode_height[opt_idx]);
    x_hat[opt_idx] = mode_height[opt_idx];
  }

  for (int j = 0; j < seq_len; ++j) {
    if (fwd_msgs[j]) {
      delete fwd_msgs[j];
      fwd_msgs[j] = NULL;
    }
    if (rev_msgs[j]) {
      delete rev_msgs[j];
      rev_msgs[j] = NULL;
    }
  }

  delete back_pointers_fwd;
  delete back_pointers_rev;

  return(R_NilValue);
}  // SEXP BinaryMatXXT_RI

}  // extern "C"
