
FLDPCheckMessage = function(msgObj) {
  numNAs = sum(sapply(msgObj, function(x) sum(is.na(x))))
  if (numNAs > 0) {
    stop("NA's in message components")
  }
}

FLDPBinomPlotMsg = function(msgObj, xlim = NULL, nPts = 100, add = FALSE, ...) {
  if (length(msgObj$x) < 2) {
    if (!add) {
      plot(c(0, 1), c(0, 1), ..., type = "n")
    }
    return(NULL)
  }

  if (is.null(xlim)) {
    xlim = c(0.5/nPts, 1 - (0.5 / nPts))
  }

  evalPts = seq(from = xlim[1], to = xlim[2], length = nPts)
  y = sapply(evalPts, FLDPBinomMsgEval, msgObj = msgObj)

  if (add) {
    lines(evalPts, y, ...)
  } else {
    plot(evalPts, y, type = "l", ...)
  }
  return(rbind(evalPts, y))
}

FLDPPlotMsg = function(msgObj, xlim = NULL, nPts = 100, add = FALSE, ...) {
  if (length(msgObj$x) < 2) {
    if (!add) {
      plot(c(0, 1), c(0, 1), ..., type = "n")
    }
    return(NULL)
  }

  if (is.null(xlim)) {
    xLwr = min(msgObj$x[msgObj$x > -Inf])
    xUpr = max(msgObj$x[msgObj$x < Inf])

    xlim = c(xLwr - 0.2 * (xUpr - xLwr),
             xUpr + 0.2 * (xUpr - xLwr))
  }

  evalPts = seq(from = xlim[1], to = xlim[2], length = nPts)
  y = sapply(evalPts, FLDPMsgEval, msgObj = msgObj)

  if (add) {
    lines(evalPts, y, ...)
  } else {
    plot(evalPts, y, type = "l", ...)
  }
  return(rbind(evalPts, y))
}

FLDPCheckContinuity = function(msgObj) {
  x = msgObj$x
  const = msgObj$const
  lin = msgObj$lin
  quad = msgObj$quad
  nSeg = length(x)

  if (nSeg == 1) {
    return(TRUE)
  } else {
    for (j in 2:nSeg) {
      v1 = x[j] * (lin[j] + x[j] * quad[j]) + const[j]
      v2 = x[j] * (lin[j-1] + x[j] * quad[j-1]) + const[j-1]
      if (abs(v1 - v2) > 1e-6) {
        stop("knots not continuous")
      }

      v1 = lin[j] + 2 * x[j] * quad[j]
      v2 = lin[j-1] + 2 * x[j] * quad[j-1]
      if (abs(v1 - v2) > 1e-6) {
        stop("knots not continuous")
      }

    }
    return(TRUE)
  }
}

FLDPCheckConcavity = function(msgObj) {
  x = msgObj$x
  const = msgObj$const
  lin = msgObj$lin
  quad = msgObj$quad
  nSeg = length(x)

  if (nSeg == 1) {
    return(TRUE)
  } else {
    lastDeriv = msgObj$lin[1]
    if (msgObj$quad[1] != 0) lastDeriv = Inf

    for (j in 2:nSeg) {
      curDeriv = lin[j] + 2 * x[j] * quad[j]

      if (curDeriv > lastDeriv + 1e-6) {
        stop("message not concave")
      } else {
        lastDeriv = curDeriv
      }
    }
    return(TRUE)
  }
}

FLDPDerivHitSub = function(msgObj, targDeriv, knotIdx) {
  return((targDeriv - msgObj$lin[knotIdx]) / (2 * msgObj$quad[knotIdx]))
}

FLDPBinomDerivHitSub = function(msgObj, targDeriv, knotIdx) {
  # code for the binomial distribution is more complicated
  lin = msgObj$lin[knotIdx]
  
  if (msgObj$log[knotIdx] == 0) {
    dnm = targDeriv - lin
    if (dnm == 0) {
      return(NULL)
    }
    r = (msgObj$log2[knotIdx] / dnm) + 1
    if (r < 0 || r > 1) {
      return(NULL)
    } else {
      return(r)
    }
  } else if (msgObj$log2[knotIdx] == 0) {
    dnm = targDeriv - lin
    if (dnm == 0) {
      return(NULL)
    }
    r = (msgObj$log[knotIdx] / dnm)
    if (r < 0 || r > 1) {
      return(NULL)
    } else {
      return(r)
    }
  } else {
    
    a = targDeriv - lin
    
    if (abs(a) < 1e-8) {
      return(msgObj$log[knotIdx] / (msgObj$log[knotIdx] + msgObj$log2[knotIdx]))
    }
    
    b = lin - targDeriv - msgObj$log[knotIdx] - msgObj$log2[knotIdx]
    c = msgObj$log[knotIdx]
    
    t1 = (b * b - 4 * a * c)
    if (t1 < 0) {
      return(NULL)
    }
    
    soln1 = (-b + sqrt(t1)) / (2 * a)
    soln2 = (-b - sqrt(t1)) / (2 * a)
    
    if (soln1 < 0 || soln1 > 1) {
      if (soln2 < 0 || soln2 > 1) stop("error finding derivative")
      return(soln2)
    } else {
      if (soln2 >= 0 && soln2 <= 1) stop("both solutions valid")
      return(soln1)
    }
  }
}

FLDPBinomDerivHit = function(msgObj, findDeriv = 0, revDir = FALSE) {
  return(FLDPDerivHit(msgObj, findDeriv = findDeriv, revDir = revDir,
                      derivHitFunc = FLDPBinomDerivHitSub, paramUB = 1))
}
                        
FLDPDerivHit = function(msgObj, findDeriv = 0, revDir = FALSE,
                        derivHitFunc = FLDPDerivHitSub, paramUB = Inf) {    
  x = msgObj$x
  nSeg = length(x)

  knotSeq = 1:nSeg
  if (revDir) knotSeq = rev(knotSeq)

  for (j in knotSeq) {
    nextKnot = ifelse (j < nSeg, x[j + 1], paramUB)
    hitPosn = derivHitFunc(msgObj, findDeriv, j)

    if (!is.null(hitPosn) && hitPosn >= x[j] && hitPosn <= nextKnot) {
      return(c(hitPosn, j))
    }
  }
  return(NULL)
}

FLDPFindKnot = function(x, msgObj) {
  knotLocs = msgObj$x   
  nSeg = length(knotLocs)

  for (j in 1:nSeg) {
    nextKnot = ifelse(j < nSeg, knotLocs[j + 1], Inf)

    if (x >= knotLocs[j] && x <= nextKnot) {
      return(j)
    }
  }
  
  return(NULL)
}

FLDPMsgEval = function(x, msgObj, knotIdx = NULL) {
  if (is.null(knotIdx)) knotIdx = FLDPFindKnot(x, msgObj)

  return(msgObj$const[knotIdx]
         + msgObj$lin[knotIdx] * x
         + msgObj$quad[knotIdx] * x * x)
}

FLDPBinomMsgEval = function(x, msgObj, knotIdx = NULL) {
  if (is.null(knotIdx)) knotIdx = FLDPFindKnot(x, msgObj)

  return(msgObj$const[knotIdx]
         + msgObj$lin[knotIdx] * x
         + msgObj$log[knotIdx] * log(x)
         + msgObj$log2[knotIdx] * log(1 - x))
}

FLDPBinomArgmax = function(lambda2, msgObj) {
  return(FLDPArgmax(lambda2, msgObj, derivFind = FLDPBinomDerivHit,
                    msgEval = FLDPBinomMsgEval,
                    otherNames = c("log", "log2"),
                    paramLB = 0, paramUB = 1))
}

FLDPArgmax = function(lambda2, msgObj, derivFind = FLDPDerivHit,
                      msgEval = FLDPMsgEval,
                      otherNames = "quad",
                      paramLB = -Inf, paramUB = Inf) {  
  leftHit = derivFind(msgObj, findDeriv = lambda2, revDir = FALSE)
  rightHit = derivFind(msgObj, findDeriv = -lambda2, revDir = TRUE)

  if (is.null(leftHit)) {
    if (is.null(rightHit)) {
      return(list(msgObj = msgObj, backPointer = c(paramLB, paramUB)))
    }
    
    rightHeight = msgEval(rightHit[1], msgObj, knotIdx = rightHit[2])
    rightConst = rightHeight + rightHit[1] * lambda2
  
    inds = 1:rightHit[2]
  
    msgObj2 = list(x = c(msgObj$x[inds], rightHit[1]),
                   const = c(msgObj$const[inds], rightConst),
                   lin = c(msgObj$lin[inds], -lambda2))
  
    for (nm in otherNames) {
      msgObj2[[nm]] = c(msgObj[[nm]][inds], 0)
    }
    
    return(list(msgObj = msgObj2, backPointer = c(paramLB, rightHit[1])))
  
  } else if (is.null(rightHit)) {
    leftHeight = msgEval(leftHit[1], msgObj, knotIdx = leftHit[2])
    leftConst = leftHeight - leftHit[1] * lambda2
  
    inds = leftHit[2]:length(msgObj$x)
    x = c(paramLB, msgObj$x[inds])
    x[2] = leftHit[1]
    
    msgObj2 = list(x = x,
                   const = c(leftConst, msgObj$const[inds]),
                   lin = c(lambda2, msgObj$lin[inds]))
  
    for (nm in otherNames) {
      msgObj2[[nm]] = c(0, msgObj[[nm]][inds])
    }
    return(list(msgObj = msgObj2, backPointer = c(leftHit[1], paramUB)))
    
  } else {
    # we should always enter this case for the Gaussian likelihood
    
    leftHeight = msgEval(leftHit[1], msgObj, knotIdx = leftHit[2])
    rightHeight = msgEval(rightHit[1], msgObj, knotIdx = rightHit[2])
    
    leftConst = leftHeight - leftHit[1] * lambda2
    rightConst = rightHeight + rightHit[1] * lambda2
  
    inds = leftHit[2]:rightHit[2]
    x = c(paramLB, msgObj$x[inds], rightHit[1])
    x[2] = leftHit[1]
  
    msgObj2 = list(x = x,
                   const = c(leftConst, msgObj$const[inds], rightConst),
                   lin = c(lambda2, msgObj$lin[inds], -lambda2))
  
    for (nm in otherNames) {
      msgObj2[[nm]] = c(0, msgObj[[nm]][inds], 0)
    }
    
    return(list(msgObj = msgObj2, backPointer = c(leftHit[1], rightHit[1])))
  }
}

FLDPBinomAddObs = function(obsVal, msgObj, obsWt = 1) {      
  msgObj$log = msgObj$log + obsVal[1] * obsWt
  msgObj$log2 = msgObj$log2 + obsVal[2] * obsWt
  return(msgObj)
}

FLDPAddObs = function(obsVal, msgObj, obsWt = 1) {
  c1 = -0.5 * obsWt
  
  return(FLDPAddQuad(const = (c1 * obsVal * obsVal),
                     lin = (-2.0 * c1 * obsVal),
                     quad = c1,
                     msgObj = msgObj))
}

FLDPAddQuad = function(const, lin, quad, msgObj) {  
  msgObj$const = msgObj$const + const  
  msgObj$lin = msgObj$lin + lin  
  msgObj$quad = msgObj$quad + quad
  return(msgObj)
}

FLDPEmptyMsg = function() {
  return(list(x = -Inf, const = 0, lin = 0, quad = 0))
}

FLDPBinomEmptyMsg = function() {
  return(list(x = 0, const = 0, lin = 0, log = 0, log2 = 0))
}

FLDPBackTrace = function(config, lastMsg, backPointers) {
  n = ncol(backPointers)

  r = rep(0, n)
  r[n] = config$derivHit(msgObj = lastMsg, findDeriv = 0)[1]

  for (j in (n-1):1) {
    if (r[j + 1] < backPointers[1, j]) {
      r[j] = backPointers[1, j]

    } else if (r[j + 1] > backPointers[2, j]) {
      r[j] = backPointers[2, j]

    } else {
      # the j'th and (j+1)'st parameters were fused
      r[j] = r[j + 1]
    }
  }
  return(r)
}

FLDPGaussConfig = function() {
  return(list(emptyMsg = FLDPEmptyMsg,
              addObs = FLDPAddObs,
              argMax = FLDPArgmax,
              derivHit = FLDPDerivHit,
              plotMsg = FLDPPlotMsg,
              getObs = function(x, idx) x[idx]))
}

FLDPBinomConfig = function() {
  return(list(emptyMsg = FLDPBinomEmptyMsg,
              addObs = FLDPBinomAddObs,
              argMax = FLDPBinomArgmax,
              derivHit = FLDPBinomDerivHit,
              plotMsg = FLDPBinomPlotMsg,
              getObs = function(x, idx) x[, idx]))
}
              

FLDPR = function(x, lambda2 = 1, lambda1 = 0,
                 maxSegs = 1000, obsWts = NULL, makePlots = TRUE,
                 config = NULL) {
  if (is.matrix(x)) {
    xLen = ncol(x)
    if (is.null(config)) {
      config = FLDPBinomConfig()
    }
    
    if (lambda1 > 0) {
      stop("lambda1 > 0 not implemented for non-Gaussian sequences")
    }
  } else {
    xLen = length(x)
    if (is.null(config)) {
      config = FLDPGaussConfig()
    }
  }
  
  if (sum(is.na(x)) > 0) {
    stop("NA's in x")
  }
  
  if (is.null(obsWts)) {
    obsWts = rep(1.0, xLen)
  }      
                                 
  nSegs = rep(as.integer(0), xLen)
  backPointers = matrix(nrow = 2, ncol = xLen, data=0)

  vitMsgs = vector(mode = "list") 
  preMsgs = vector(mode = "list")
  
  ##############################################################################
  # the forward pass
  
  curMsg = config$emptyMsg()
  
  for (j in 1:xLen) {
    preMsgs[[j]] = config$addObs(obsVal = config$getObs(x, j), obsWt = obsWts[j],
                                 msgObj = curMsg)
    r = config$argMax(lambda2, msgObj = preMsgs[[j]])

    vitMsgs[[j]] = r$msgObj
    backPointers[, j] = r$backPointer
  
    nSegs[j] = length(vitMsgs[[j]]$x)

    FLDPCheckMessage(preMsgs[[j]])
    FLDPCheckMessage(vitMsgs[[j]])
    
    if (j == xLen) {
      curMsg = preMsgs[[j]]
    } else {
      curMsg = vitMsgs[[j]]
    }

    if (makePlots) {
      config$plotMsg(msgObj = preMsgs[[j]], nPts = 100, add = FALSE)
      config$plotMsg(msgObj = vitMsgs[[j]], nPts = 100, add = TRUE, col = "red")
      scan()
    }
  }
  
  ##############################################################################
  # perform the back-trace
  backTrace = FLDPBackTrace(config, curMsg, backPointers)

  ##############################################################################
  # Softhreshold if 'lambda1' is positive.  This will not be useful or correct
  # for the binomial model.
  softthresh = function(x, amt) {
    (x-sign(x)*amt)*(sign(x) == sign(x-sign(x)*amt) )
  }
  backTrace = softthresh(backTrace, lambda1)

  return(list(fit = backTrace, preMsgs = preMsgs, vitMsgs = vitMsgs,
              backPointers = backPointers))
}
