library("genlasso")
library("l1tf")
#-------------
# main function to run
# burst.binom - getting the regularized estimation for binomial time series
# k, N - vectors of same length T, 
#        N is the total number of samples at each time, 
#        k is the number of samples that are positive
# lambda - regularization parameter
# ord - when ord=0, penalize the 1st order difference (fused lasso, l0-segmentation)
#       when ord=1, penalize the 2nd order difference (trend-filtering)
# use.l0 - if ord=0, then do l0 segmentation if set to TRUE. Otherwise do the l1-penalty
# maxIter - max number of iteration
# tol - tolerance
# p.base - a baseline probability used to identify bursts
# p.init - initial value of the optimization
# silent - no print while fitting
# plot.proc - plot the fitting process
# post.trunc - (recommended to be TRUE), if FALSE, will use a baseline during the fitting rather than after the fitting.
#-------------

burst.binom <- function(k, N, lambda = 200,
                        maxIter = 2000, tol = 1e-9,
                        p.base = sum(k) / sum(N),
                        p.init = sum(k) / sum(N),
                        silent = FALSE,
                        use.l0 = FALSE,
                        plot.proc = FALSE,
                        post.trunc = TRUE,
                        ord = 0){
  if(post.trunc){
    p.base.use = 0
  } else {
    p.base.use = p.base
  }
  if(use.l0){
    if(ord==0){
      rlt <- flasso.binom.ista.pos(k, N, lambda=lambda, maxIter=maxIter, tol=tol, 
                                 p.base=p.base.use, p.init=p.init, plot.proc=plot.proc, 
                                 use.l0=use.l0, silent=silent)
    } else {
      print("not implemented yet")
      return()
    }
  } else {
      rlt <- flasso.binom.fista.pos(k, N, lambda=lambda, maxIter=maxIter, tol=tol, 
                                    p.base=p.base.use, p.init=p.init, plot.proc=plot.proc, 
                                    ord=ord, silent=silent)
  }
  
  #post-processing, identify bursts
  p.hat <- rlt$p.hat
  theta.hat <- rlt$theta.hat
  theta.base <- log(p.base / (1 - p.base))
  is.burst <- p.hat > (p.base + 1e-6) #in case of 
  dsig <- diff(c(0, is.burst, 0));
  ind.start <- which(dsig > 0);
  ind.end <- which(dsig < 0) - 1;
  nBurst <- length(ind.start)
  weight <- numeric(nBurst)
  l0 <- numeric(nBurst); l1 <- l0; pen <- l0;
  
  if(nBurst > 0){
    for(j in 1:nBurst){
      b.range <- ind.start[j]:ind.end[j]
      k.use <- k[b.range]
      N.use <- N[b.range]
      theta.use <- theta.hat[b.range]
      l0[j] <- -sum(k.use) * theta.base + sum(N.use) * log(1 + exp(theta.base))
      l1[j] <- -sum(k.use * theta.use) + sum(N.use * log(1 + exp(theta.use)))
    }
    weight <- l0 - l1# - pen
  }
  
  bursts <- data.frame(start = ind.start, end = ind.end, weight = weight,
                       l0 = l0, l1 = l1)  
  rlt$bursts <- bursts
  rlt$p.base <- p.base
  return(rlt)
}

#--------------
# FISTA: used for fused-lasso
#--------------
flasso.binom.fista.pos <- function(k, N, lambda = 200,
                                   maxIter = 2000, tol = 1e-9,
                                   p.base = sum(k) / sum(N),
                                   p.init = sum(k) / sum(N), ord=0,
                                   plot.proc = FALSE, silent = TRUE){
  #used fista to solve fused lasso with binomial likelihood
  nDay <- length(k)
  theta.base <- log(p.base / (1 - p.base)) #baseline
  theta.init <- log(p.init / (1 - p.init))
  if(length(theta.init) == 1){
    theta.init <- rep(theta.init, nDay)
  }
  tSeq <- numeric(maxIter + 1)
  
  L <- max(N) / 4
  theta.prior <- theta.init
  theta.now <- theta.init
  
  #handle special case
  if(max(k) == 0){
    theta.now <- numeric(length(k)) + theta.base
    p.now <- numeric(length(k)) + p.base
    #bursts <- NULL
    return(list(theta.hat = theta.now, 
                p.hat = p.now))
  }
  
  #FISTA
  for(i in 1:maxIter){  
    tSeq[i + 1] <- (1 + sqrt(1 + 4 * tSeq[i]^2)) / 2
    y <- theta.now + (tSeq[i] - 1) / tSeq[i + 1] * (theta.now - theta.prior)
    p <- exp(y) / (1 + exp(y))
    y.use <- y - (N * p - k) / L
    theta.prior <- theta.now
    if(ord == 0){
      theta.now <- FlsaDp(y.use, lambda2 = lambda / L)
    } else {
      theta.now <- l1tf(y.use, lambda = lambda/L)
    }
    theta.now <- pmax(theta.now, theta.base)
    if(max((theta.now - theta.prior)^2) < tol){
      if(!silent) print(paste("tolerance attained after", i, "iteration!"))
      break
    } else {
      if(i == maxIter){
        if(!silent) print(paste("After", i, "iteration, convergence not attained!"))
      }
    }
    
    if(plot.proc){
      #print(paste(i, "iteration done!"))
      if(i %% 10 == 1){
        par(mfrow = c(2,1))
        plot(k, ylab = "num of relevant cables", xlab = "day")
        plot(theta.now, main = paste(i, "iter"), xlab = "day")
        Sys.sleep(1e-7)
      }
    }
  }
  p.now <- exp(theta.now) / (1 + exp(theta.now))
  
  return(list(theta.hat = theta.now, 
              p.hat = p.now))
}


#--------------------
# ISTA: suitable for L0 segmentation
#--------------------
flasso.binom.ista.pos <- function(k, N, lambda = 200, 
                                    maxIter = 5000, tol = 1e-9,
                                    p.base = 0,
                                    p.init = sum(k) / sum(N),
                                    plot.proc = FALSE,
                                    use.l0 = TRUE,
                                    silent = TRUE){
  nDay <- length(k)
  theta.init <- log(p.init / (1 - p.init))
  if(length(theta.init) == 1){
    theta.init <- rep(theta.init, nDay)
  }
  theta.base <- log(p.base / (1 - p.base)) #baseline
  L <- N / 4
  theta.prior <- theta.init
  theta.now <- theta.init
  
  #handle special case
  if(max(k) == 0){
    theta.now <- numeric(length(k)) + theta.base
    p.now <- numeric(length(k)) + p.base
    #bursts <- NULL
    return(list(theta.hat = theta.now, 
                p.hat = p.now))
  }
  
  for(i in 1:maxIter){
    p <- exp(theta.now) / (1 + exp(theta.now))
    y.use <- theta.now - (N * p - k) / L
    theta.prior <- theta.now
    if(use.l0)
      theta.now <- L0Gauss(y.use, lambda2 = lambda, obsWts = L)
    else
      theta.now <- FlsaDp(y.use, lambda2 = lambda, obsWts = L)
    theta.now <- pmax(theta.now, theta.base)
    if(max((theta.now - theta.prior)^2) < tol){
      if(!silent) print(paste("tolerance attained after", i, "iteration!"))
      break
    }    
    #if(i %% 100 == 1){
    #  print(paste(i, "iteration done!"))
    #}    
    if(plot.proc){
      #print(paste(i, "iteration done!"))
      if(i %% 1 == 0){
        Sys.sleep(0.1)
        par(mfrow = c(2,1))
        plot(k, ylab = "num of relevant cables", xlab = "day")
        plot(theta.now, main = paste(i, "iter"), xlab = "day")
        Sys.sleep(0.1)
      }
    }
  }
  if(i == maxIter){
    if(!silent) print(paste("After", i, "iteration, convergence not attained!"))
  }
  p.now <- exp(theta.now) / (1 + exp(theta.now))
  return(list(theta.hat = theta.now, 
              p.hat = p.now)) 
}

#---------------------
# cross-validation
#---------------------

#choosing optimal parameter of fused lasso through cross-validation
burst.binom.cv = function(k, N, 
                 maxIter = 2000, tol = 1e-9,
                 p.base = sum(k) / sum(N), 
                 p.init = sum(k) / sum(N),
                 nFold = 10,
                 use.l0 = FALSE,
                 post.trunc = TRUE,
                 lambda.use = c("lambda.min","lambda.1se"),
                 lambda.seq = exp(seq(log(2000), log(1), length.out = 30)),
                 ord=0){
  lambda.seq <- lambda.seq
  
  nLambda <- length(lambda.seq)
  
  kOrig = k
  NOrig = N
  
  k = k[1:(floor(length(N)/nFold)*nFold)]
  N = N[1:(floor(length(N)/nFold)*nFold)]
  
  nDay <- length(k)
  CVdf = data.frame('Data' = k, 'Fold' = rep_len(1:nFold,length(N)))
  
  log.lik.mat <- matrix(ncol = nFold, nrow = nLambda)
  for(i in 1:nFold){
    trainData = CVdf$Data[which(CVdf$Fold != i)]
    theta.train <- burst.binom.seq(trainData, N[which(CVdf$Fold != i)], maxIter = maxIter, 
                                   lambda.seq = lambda.seq, tol = tol,
                                   p.base = p.base, p.init = p.init,
                                   use.l0 = use.l0,
                                   post.trunc = post.trunc, ord=ord)
    theta.valid = array(0,dim=c(length(N),length(lambda.seq)))
    theta.valid[which(CVdf$Fold != i),] <- theta.train
    theta.valid = rbind(theta.valid[2,],theta.valid)[which(CVdf$Fold == i),]

    p.valid <- 1 - 1 / (1 + exp(theta.valid))
    
    testData = CVdf$Data[which(CVdf$Fold == i)]
    log.lik.temp <- testData * log(p.valid) + (N[which(CVdf$Fold == i)] - testData) * log(1 - p.valid)
    log.lik.temp[is.nan(log.lik.temp)] <- 0
    log.lik.mat[,i] <- as.vector(colSums(log.lik.temp)) / sum(N[which(CVdf$Fold == i)])
    
  }
  err.mean <- - rowMeans(log.lik.mat)
  err.se <- apply(log.lik.mat, 1, sd) / sqrt(nFold)
  i.min <- which.min(err.mean)
  i.1se <- which.max(err.mean < (err.mean[i.min] + err.se[i.min]))
  lambda.min <- lambda.seq[i.min]
  lambda.1se <- lambda.seq[i.1se]
  if(lambda.use[1] == "lambda.min"){
    lambda.opt <- lambda.min
  } else {
    lambda.opt <- lambda.1se
  }
  burst.binom.rlt <- burst.binom(kOrig, NOrig, lambda = lambda.opt,
                                 maxIter = maxIter, tol = tol,
                                 p.base = p.base, p.init = p.init,
                                 use.l0 = use.l0,
                                 post.trunc = post.trunc,
                                 silent = TRUE, ord=ord)
  return(list(burst.binom.rlt = burst.binom.rlt,
              lambda.opt = lambda.opt,
              err.mean = err.mean,
              err.se = err.se,
              lambda.min = lambda.min,
              lambda.1se = lambda.1se,
              lambda.seq = lambda.seq))
}


#---------------
# ancillary function
#--------------
burst.binom.seq <- function(k, N, lambda.seq = NULL,
                             maxIter = 2000,  tol = 1e-9,
                             p.base = sum(k) / sum(N),
                             p.init = p.base,
                             use.l0 = FALSE,
                             post.trunc = TRUE,
                             ord = 0){
  if(length(lambda.seq) == 0)
    lambda.seq <- exp(seq(log(500), 0, length.out = 50))
  nLambda <- length(lambda.seq)
  nDay <- length(k)
  theta.hat.mat <- matrix(nrow = nDay, ncol = nLambda)
  theta0 <- log(p.base / (1 - p.base))
  for(i in 1:nLambda){
    lambda.use <- lambda.seq[i]
    flasso.rlt <- burst.binom(k, N, lambda = lambda.use,
                              maxIter = maxIter, tol = tol,
                              p.base = p.base, p.init = p.init, 
                              plot.proc = FALSE, use.l0 = use.l0,
                              post.trunc = post.trunc, silent = TRUE,
                              ord=ord)
    theta.hat.mat[,i] <- flasso.rlt$theta.hat
    p.init <- 1 - 1 / (1 + exp(flasso.rlt$theta.hat))
  }
  if(post.trunc){
    theta.base <- log(p.base / (1-p.base))
    theta.hat.mat <- pmax(theta.hat.mat, theta.base)
  }
  return(theta.hat.mat)
}
